import re
import os
import shutil
from datetime import datetime
import uuid
import subprocess
import argparse
from string import Template

BASE_DIR = "automated_jobs"
ARCHIVE_DIR = "Archive"
LOG_FILE = os.path.join(ARCHIVE_DIR, "yaml_log.txt")

os.makedirs(ARCHIVE_DIR, exist_ok=True)


def create_data_dirs(dirs):
    return "\n" + "\n".join([f"- {d}/images" for d in dirs])


def write_yaml(filename, parameters, template_name):
    with open(f"templates/{template_name}.yaml", "r") as file:
        content = file.read()

    content = Template(content)
    filled_content = content.substitute(parameters)

    with open(filename, "w") as file:
        file.write(filled_content)


def run_job(filename, name=None):
    command = ["az", "ml", "job", "create", "-f", filename]
    if name is not None:
        command.append("--name")
        command.append(name)

    print(f"Executing: {' '.join(command)}")

    try:
        result = subprocess.run(command, capture_output=True, text=True, check=True)
        print(f"Output: {result.stdout}")
    except subprocess.CalledProcessError as e:
        print(f"Error: {e.stderr}")


def archive_yaml(filename):
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    tmp_name = f"{os.path.basename(filename).split('.')[0]}_{timestamp}.yaml"
    archive_filename = os.path.join(ARCHIVE_DIR, tmp_name)
    shutil.move(filename, archive_filename)

    timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    with open(LOG_FILE, "a") as log:
        log.write(f"{timestamp}: Archived {archive_filename}\n")
        with open(archive_filename, "r") as file:
            content = file.read()
            log.write(f"{content}\n\n\n")


def azure_path_format(path):
    path = path.strip()
    pattern = r"^azureml:[^:]+:\d+$"
    if re.match(pattern, path):
        return path
    pattern = r"^[^:]+:\d+$"
    if re.match(pattern, path):
        return f"azureml:{path}"
    return f"azureml:{path}:1"


def main():
    parser = argparse.ArgumentParser(description="Process a YAML file and run a job.")
    parser.add_argument(
        "--job", type=str, required=True, help="Type of job to run: `train` or `val`"
    )
    parser.add_argument("--job-name", type=str, help="Name of job")
    parser.add_argument(
        "--exp-name", type=str, required=True, help="Name of experiment"
    )
    parser.add_argument(
        "--train-data",
        type=str,
        default="azureml:agnostic-dupeless-fishonly:1",
        help="Path to Azure training data; default=azureml:agnostic-dupeless-fishonly:1",
    )
    parser.add_argument(
        "--pre-model",
        type=str,
        default="azureml:yolov8m-vas:1",
        help="Path to Azure pretrained model; default=azureml:yolov8m-vas:1",
    )
    parser.add_argument(
        "--epochs",
        type=int,
        default=50,
        help="Number of Epochs the model should be trained for; default=50",
    )
    parser.add_argument(
        "--batch",
        type=int,
        default=16,
        help="Batch size of validation; default=16",
    )
    parser.add_argument(
        "--iou",
        type=int,
        default=0.6,
        help="IOU of validation; default=0.6",
    )
    parser.add_argument(
        "--env",
        type=str,
        default="azureml:yolov8-environment:1",
        help="Azure environment used for the job; default=azureml:yolov8-environment:1",
    )
    parser.add_argument(
        "--compute",
        type=str,
        default="azureml:gpucluster",
        help="Name of compute that will be used for the job; default=azureml:gpucluster",
    )
    parser.add_argument(
        "--train-dirs",
        nargs="+",
        default=["orpc1", "voith1", "wells1"],
        help=(
            "Which data sets to include in the training data.",
            "Default is `--train-dirs orpc1 voith1 wells1`",
        ),
    )
    parser.add_argument(
        "--base-data-dir",
        type=str,
        default="data",#may need a slash behind it
        help=(
            "Name of base data directory, where all the train-dir and val-dir folders are located",
            "Usual value is `data`",
        )
    )
    parser.add_argument(
        "--val-dirs",
        nargs="+",
        default=["orpc2", "voith2", "wells2"],
        help=(
            "Which data sets to include in the validation/testing data.",
            "Default is `--val-dirs orpc2 voith2 wells2`",
        ),
    )

    args = parser.parse_args()

    # Ensure the job type is valid
    job_error = "--job argument must be `train` or `val`"
    assert args.job in ["train", "val"], job_error

    # Ensure arzure path arguments have the correct format
    args.train_data = azure_path_format(args.train_data)
    args.pre_model = azure_path_format(args.pre_model)
    args.env = azure_path_format(args.env)

    # Convert numbers to strings
    args.epochs = str(args.epochs)
    args.batch = str(args.batch)
    args.iou = str(args.iou)

    if args.job == "train":
        command = ' |\n  sed -i "s|path:.*$|path: ${{inputs.training_data}}|" ${{inputs.dataset_yaml}}\n  yolo task=detect train data=${{inputs.dataset_yaml}} model=${{inputs.model_to_train}} epochs=${{inputs.epochs}} project=${{inputs.experiment_name}} name=experiment'
    elif args.job == "val":
        command = ' |\n  sed -i "s|path:.*$|path: ${{inputs.training_data}}|" ${{inputs.dataset_yaml}}\n  yolo task=detect val data=${{inputs.dataset_yaml}} model=${{inputs.model_to_train}} batch=${{inputs.batch}} iou=${{inputs.iou}} project=${{inputs.experiment_name}} name=experiment'

    train_dirs = create_data_dirs(args.train_dirs)
    val_dirs = create_data_dirs(args.val_dirs)
    base_data_directory = args.base_data_dir

    unique_id = str(uuid.uuid4())[:4]

    data_yaml_filename = f"data_{unique_id}.yaml"
    data_yaml_path = os.path.join(BASE_DIR, data_yaml_filename)
    data_yaml_parameters = {
        "train_dirs": train_dirs,
        "val_dirs": val_dirs,
        "base_data_dir": base_data_directory
    }
    write_yaml(data_yaml_path, data_yaml_parameters, "data")

    job_yaml_filename = f"job_{unique_id}.yaml"
    job_yaml_path = os.path.join(BASE_DIR, job_yaml_filename)
    job_yaml_parameters = {
        "command": command,
        "exp_name": args.exp_name,
        "train_data": args.train_data,
        "pre_model": args.pre_model,
        "data_yaml": data_yaml_filename,
        "epochs": args.epochs,
        "batch": args.batch,
        "iou": args.iou,
        "env": args.env,
        "compute": args.compute,
    }
    write_yaml(job_yaml_path, job_yaml_parameters, "job")
    # quit()
    run_job(job_yaml_path, args.job_name)
    archive_yaml(job_yaml_path)
    archive_yaml(data_yaml_path)


if __name__ == "__main__":
    main()