"""
create_fish_and_fishless_data.py creates a unique data directory.
It assumes the raw data is at the given paths of IN_* as referenced in
the variables below from the src.data_paths.orginal module.

This script is the most basic of the sorting scripts:
- removes duplicates
- contains all fish images, a 20% fishless, and 30% fishless folder of images
- creates an agnostic structure for out data that is divided by contiguous stream of video

data/
├── orpc1
│   ├── fish
│   │   ├── images
│   │   └── labels
│   ├── fishless_20percent
│   │   └── images
│   └── fishless_30percent
│       └── images
├── orpc2
│   ├── fish
│   │   ├── images
│   │   └── labels
│   ├── fishless_20percent
│   │   └── images
│   └── fishless_30percent
│       └── images
├── orpc3
│   ├── fish
│   │   ├── images
│   │   └── labels
│   ├── fishless_20percent
│   │   └── images
│   └── fishless_30percent
│       └── images
├── orpc4
│   ├── fish
│   │   ├── images
│   │   └── labels
│   ├── fishless_20percent
│   │   └── images
│   └── fishless_30percent
│       └── images
├── orpc5
│   ├── fish
│   │   ├── images
│   │   └── labels
│   ├── fishless_20percent
│   │   └── images
│   └── fishless_30percent
│       └── images
├── voith1
│   ├── fish
│   │   ├── images
│   │   └── labels
│   ├── fish_1-100
│   │   ├── images
│   │   └── labels
│   ├── fish_101-200
│   │   ├── images
│   │   └── labels
│   ├── fish_201-300
│   │   ├── images
│   │   └── labels
│   ├── fishless_20percent
│   │   └── images
│   └── fishless_30percent
│       └── images
├── voith2
│   ├── fish
│   │   ├── images
│   │   └── labels
│   ├── fishless_20percent
│   │   └── images
│   └── fishless_30percent
│       └── images
├── voith3
│   ├── fish
│   │   ├── images
│   │   └── labels
│   ├── fish_1-100
│   │   ├── images
│   │   └── labels
│   ├── fish_101-200
│   │   ├── images
│   │   └── labels
│   ├── fish_201-300
│   │   ├── images
│   │   └── labels
│   ├── fishless_20percent
│   │   └── images
│   └── fishless_30percent
│       └── images
├── voith4
│   ├── fish
│   │   ├── images
│   │   └── labels
│   ├── fishless_20percent
│   │   └── images
│   └── fishless_30percent
│       └── images
├── wells1
│   ├── fish
│   │   ├── images
│   │   └── labels
│   ├── fish_1-100
│   │   ├── images
│   │   └── labels
│   ├── fish_101-200
│   │   ├── images
│   │   └── labels
│   ├── fish_201-300
│   │   ├── images
│   │   └── labels
│   ├── fishless_20percent
│   │   └── images
│   └── fishless_30percent
│       └── images
└── wells2
    ├── fish
    │   ├── images
    │   └── labels
    ├── fish_1-100
    │   ├── images
    │   └── labels
    ├── fish_101-200
    │   ├── images
    │   └── labels
    ├── fish_201-300
    │   ├── images
    │   └── labels
    ├── fishless_20percent
    │   └── images
    └── fishless_30percent
        └── images
"""

import os
import random
import shutil
import cv2
import numpy as np
from src.data import get_files, nat_sort, write_label
from src.utils import loading_bar
from src.yolo_utils import get_object_strings
from src.data_paths.raw import (
    IN_TRAIN_ORPC1,
    IN_TRAIN_ORPC2,
    IN_TRAIN_ORPC3,
    IN_TRAIN_VOITH,
    IN_TRAIN_WELLS,
    IN_TEST_ORPC1,
    IN_TEST_ORPC2,
    IN_TEST_VOITH1,
    IN_TEST_VOITH2,
    IN_TEST_VOITH3,
    IN_TEST_WELLS,
)
from src.data_paths.exp_303 import (
    OUT_ORPC1,
    OUT_ORPC2,
    OUT_ORPC3,
    OUT_VOITH1,
    OUT_WELLS1,
    OUT_ORPC4,
    OUT_ORPC5,
    OUT_VOITH2,
    OUT_VOITH3,
    OUT_VOITH4,
    OUT_WELLS2,
)


def process_data(pattern, dest_dir, remove_dupes=True):
    """Copies all images and labels from pattern and puts them in the destination directory.

    Args:
        pattern (str): Pattern to find all source data.
        dest_dir (str): Destination directory to copy data to.
        remove_dupes (bool): Does not save duplicate images if True. Default is True.
    """
    if dest_dir is None:
        return
    print(f"Searching {pattern} and outputing to {dest_dir}.")
    image_paths, label_paths = get_files(pattern)
    n = len(list(image_paths))

    dest_fish_dir = os.path.join(dest_dir, "fish")
    dest_fish_image_dir = os.path.join(dest_fish_dir, "images")
    dest_fish_label_dir = os.path.join(dest_fish_dir, "labels")
    dest_fishless_image_dir = os.path.join(dest_dir, "fishless/images")
    os.makedirs(dest_fish_image_dir, exist_ok=True)
    os.makedirs(dest_fish_label_dir, exist_ok=True)
    os.makedirs(dest_fishless_image_dir, exist_ok=True)

    last_image = cv2.imread(image_paths[-1])
    counts = {
        "fish_detections": 0,
        "fish_images": 0,
        "fishless_images": 0,
        "fish_image_dupes": 0,
        "fishless_image_dupes": 0,
    }
    for i in range(n):
        image = cv2.imread(image_paths[i])
        if remove_dupes and np.array_equal(image, last_image):
            if len(get_object_strings(label_paths[i])) > 0:
                counts["fish_image_dupes"] += 1
            else:
                counts["fishless_image_dupes"] += 1
            last_image = image
            loading_bar(i, n)
            continue
        last_image = image

        object_strings = get_object_strings(label_paths[i])
        if len(object_strings) > 0:
            image_name = os.path.basename(image_paths[i])
            label_name = os.path.basename(label_paths[i])
            label_name = os.path.splitext(image_name)[0] + ".txt"

            dest_image_path = os.path.join(dest_fish_image_dir, image_name)
            dest_label_path = os.path.join(dest_fish_label_dir, label_name)

            shutil.copy(image_paths[i], dest_image_path)
            write_label(dest_label_path, object_strings)

            counts["fish_detections"] += len(object_strings)
            counts["fish_images"] += 1
        else:
            image_name = os.path.basename(image_paths[i])

            dest_image_path = os.path.join(dest_fishless_image_dir, image_name)

            shutil.copy(image_paths[i], dest_image_path)

            counts["fishless_images"] += 1

        loading_bar(i, n)

    if pattern == IN_TRAIN_VOITH:
        counts = break_up_fish_dir(
            dest_dir, dest_fish_dir, counts, breaks=[100, 200, 300]
        )
    if pattern == IN_TRAIN_WELLS:
        counts = break_up_fish_dir(
            dest_dir, dest_fish_dir, counts, breaks=[9384, 11260]
        )
    if pattern == IN_TEST_VOITH2:
        counts = break_up_fish_dir(
            dest_dir, dest_fish_dir, counts, breaks=[100, 200, 300]
        )
    if pattern == IN_TEST_WELLS:
        counts = break_up_fish_dir(
            dest_dir, dest_fish_dir, counts, breaks=[806, 2145]
        )

    counts = break_up_fishless_dir(dest_dir, dest_fishless_image_dir, counts)

    print(f"{counts=}")


def break_up_fish_dir(dest_dir, src, counts, breaks=None, remove=False):
    """Breaks up the fish image directory with the given break structure.

    Args:
        dest_dir (string): Directory where the sub directories will be placed.
        src (string): Directory where the images/ and labels/ directories are found to copy from.
        counts (dict): Dictionary of all current image counts.
        breaks (list, optional): List of how the dest_dir should be broken up. Defaults to None.
        remove (bool, optional): Remove the original file after copying it. Defaults to True.

    Returns:
        dict: Dictionary of image counts.
    """
    if breaks is None:
        breaks = [100, 200, 300]

    src_images = os.path.join(src, "images")
    fish_image_paths = [os.path.join(src_images, f) for f in os.listdir(src_images)]
    fish_image_paths = nat_sort(fish_image_paths)
    fish_label_paths = []
    for path in fish_image_paths:
        label_path = path.replace("/images/", "/labels/")
        label_path = label_path.replace(".png", ".txt")
        fish_label_paths.append(label_path)

    if breaks[-1] > len(fish_image_paths):
        print(
            f"SKIPPING: End break {breaks[-1]} must be within the number of images available {len(fish_image_paths)}."
        )
        return counts

    start_idx = 1
    for end_idx in breaks:
        index_name = f"fish_{start_idx}-{end_idx}"
        dest_fish_image_dir = os.path.join(dest_dir, f"{index_name}/images")
        dest_fish_label_dir = os.path.join(dest_dir, f"{index_name}/labels")
        os.makedirs(dest_fish_image_dir, exist_ok=True)
        os.makedirs(dest_fish_label_dir, exist_ok=True)

        counts[index_name] = end_idx - start_idx + 1

        for i in range(start_idx - 1, end_idx):
            src_image_path = fish_image_paths[i]
            image_name = os.path.basename(src_image_path)
            dest_image_path = os.path.join(dest_fish_image_dir, image_name)
            shutil.copy(src_image_path, dest_image_path)

            src_label_path = fish_label_paths[i]
            label_name = os.path.basename(src_label_path)
            dest_label_path = os.path.join(dest_fish_label_dir, label_name)
            shutil.copy(src_label_path, dest_label_path)

            if remove:
                shutil.rmtree(src_image_path)
                shutil.rmtree(src_label_path)

            loading_bar(i - start_idx - 1, counts[index_name])

        start_idx = end_idx + 1

    return counts


def break_up_fishless_dir(dest_dir, src, counts, remove=True):
    """Breaks up the fishless image directory into 20% and 30% Fish:Fishless ratios.

    Args:
        dest_dir (string): Directory where the sub directories will be placed.
        src (string): Directory where the images/ directory is found to copy from.
        counts (dict): Dictionary of all current image counts.
        remove (bool, optional): Remove the src directory once files are copied. Defaults to True.

    Returns:
        dict: Dictionary of image counts.
    """
    dest_fishless20_image_dir = os.path.join(dest_dir, "fishless_20percent/images")
    dest_fishless30_image_dir = os.path.join(dest_dir, "fishless_30percent/images")
    os.makedirs(dest_fishless20_image_dir, exist_ok=True)
    os.makedirs(dest_fishless30_image_dir, exist_ok=True)

    num_perc = lambda p, y: int((p * y) / (1 - p))
    counts["fishless_20percent_images"] = num_perc(0.2, counts["fish_images"])
    counts["fishless_30percent_images"] = num_perc(0.3, counts["fish_images"])

    fishless_image_paths = [os.path.join(src, f) for f in os.listdir(src)]

    n = counts["fishless_20percent_images"] + counts["fishless_30percent_images"]
    try:
        random_fishless_paths = random.sample(fishless_image_paths, n)
    except:
        print("Not enough fishless images to fill both fishless directories.")
        n = min(counts["fishless_20percent_images"], len(fishless_image_paths))
        print(f"{dest_fishless20_image_dir} will have {n} fishless images.")
        n = len(fishless_image_paths) - counts["fishless_20percent_images"]
        print(f"{dest_fishless30_image_dir} will have {n} fishless images.")
        random.shuffle(fishless_image_paths)
        random_fishless_paths = fishless_image_paths

    n = counts["fishless_20percent_images"]
    for i, src_image_path in enumerate(random_fishless_paths[:n]):
        image_name = os.path.basename(src_image_path)
        dest_image_path = os.path.join(dest_fishless20_image_dir, image_name)
        shutil.copy(src_image_path, dest_image_path)
        loading_bar(i, n)

    n_30 = len(random_fishless_paths[n:])
    for i, src_image_path in enumerate(random_fishless_paths[n:]):
        image_name = os.path.basename(src_image_path)
        dest_image_path = os.path.join(dest_fishless30_image_dir, image_name)
        shutil.copy(src_image_path, dest_image_path)
        loading_bar(i, n_30)

    if remove:
        shutil.rmtree(os.path.dirname(src))

    return counts


if __name__ == "__main__":
    process_data(IN_TRAIN_ORPC1, OUT_ORPC1)
    process_data(IN_TRAIN_ORPC2, OUT_ORPC2)
    process_data(IN_TRAIN_ORPC3, OUT_ORPC3)
    process_data(IN_TRAIN_VOITH, OUT_VOITH1)
    process_data(IN_TRAIN_WELLS, OUT_WELLS1)
    process_data(IN_TEST_ORPC1, OUT_ORPC4)
    process_data(IN_TEST_ORPC2, OUT_ORPC5)
    process_data(IN_TEST_VOITH1, OUT_VOITH2)
    process_data(IN_TEST_VOITH2, OUT_VOITH3)
    process_data(IN_TEST_VOITH3, OUT_VOITH4)
    process_data(IN_TEST_WELLS, OUT_WELLS2)