import os
import cv2
import xml.etree.ElementTree as ET

import numpy as np


def get_object_strings(path):
    """Reads an xml file that encodes labeled objects.

    Args:
        path (str): Path to xml file.

    Returns:
        list: List of strings that encodes the labeled object in a format that YOLO can understand.
    """
    tree = ET.parse(path)
    root = tree.getroot()

    width = int(root.find("size").find("width").text)
    height = int(root.find("size").find("height").text)

    # extract objects
    objects = root.findall("object")
    object_strings = []
    for detection in objects:
        bndbox = detection.find("bndbox")
        xmin = int(bndbox.find("xmin").text)
        xmax = int(bndbox.find("xmax").text)
        ymin = int(bndbox.find("ymin").text)
        ymax = int(bndbox.find("ymax").text)

        # Convert from min/max x/y to normalized center, width, height
        object_center_x = (xmin + xmax) / 2 / width
        object_center_y = (ymin + ymax) / 2 / height
        normalized_object_width = (xmax - xmin) / width
        normalized_object_height = (ymax - ymin) / height
        object_strings.append(
            f"0 {object_center_x} {object_center_y} {normalized_object_width} {normalized_object_height}"
        )
    return object_strings


def draw_boxes(image, label_path):
    """Draws a box using the labels from the given path

    Args:
        image (numpy.ndarray): Numpy array of the image.
        label_path (str): Path to the corresponding label of the image.

    Returns:
        numpy.ndarray: Original image with added pixels that box the object labeled in the image.
    """
    base_name = os.path.basename(label_path)
    base_name = os.path.splitext(base_name)[0]
    cv2.putText(
        image,
        base_name,
        (10, image.shape[1] - 120),
        cv2.FONT_HERSHEY_SIMPLEX,
        0.5,
        (0, 0, 255),
        1,
        cv2.LINE_AA,
    )
    if not os.path.exists(label_path):
        return image

    detections = get_object_strings(label_path)
    if not detections:
        return image

    height, width = image.shape[:2]

    for detection in detections:
        label_data = detection.split(" ")
        class_id, x_center, y_center, box_width, box_height = map(float, label_data)
        x_center *= width
        y_center *= height
        box_width *= width
        box_height *= height
        x_min = int(x_center - (box_width / 2))
        y_min = int(y_center - (box_height / 2))
        x_max = int(x_center + (box_width / 2))
        y_max = int(y_center + (box_height / 2))
        color = (255, 0, 0)
        cv2.rectangle(image, (x_min, y_min), (x_max, y_max), color, 1)
        cv2.putText(
            image,
            "Fish",
            (x_min, y_min - 10),
            cv2.FONT_HERSHEY_SIMPLEX,
            0.3,
            (0, 255, 0),
            1,
            cv2.LINE_AA,
        )
    return image


def apply_clahe(image, clipLimit=2.0, tileGridSize=None):
    """Applys a clahe filter to the image

    Args:
        image (numpy.ndarray): Original image.
        clipLimit (float): Limits the histogram range. Default is 2.0.
        tileGridSize (tuple(int, int)): Grid to divide the image in. Default is (8, 8).

    Returns:
        numpy.ndarray: Image with CLAHE applied.
    """
    if tileGridSize is None:
        tileGridSize = (8, 8)
    clahe = cv2.createCLAHE(clipLimit=clipLimit, tileGridSize=tileGridSize)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2LAB)
    l, a, b = cv2.split(image)
    clahe_l = clahe.apply(l)
    lab_clahe = cv2.merge((clahe_l, a, b))
    return cv2.cvtColor(lab_clahe, cv2.COLOR_LAB2BGR)


def compute_image_difference(img1, img2):
    """Returns an image that represents the difference between two images.

    Args:
        img1 (numpy.ndarray): First image.
        img2 (numpy.ndarray): Second image.

    Returns:
        numpy.ndarray: Difference between images.
    """
    assert img1.shape == img2.shape, "Image dimensions don't match."
    diff = np.abs(img1.astype(np.int16) - img2.astype(np.int16))
    return np.clip(diff, 0, 255).astype(np.uint8)